
/* Define as we already include the implementation in the header (to avoid
redefinition*/
#ifndef RMMATRIX2D_C
#define RMMATRIX2D_C

#include <libgen.h>
#include <stdio.h>

#include "rmmatrix2d.h"
#include "caract/rmlib/rmutils/rmutils.h"

template<class T>
rmMatrix2d<T>::rmMatrix2d()
{
	data = NULL;
	dimX=dimY=0;
	orgX=orgY=0;
	resize(0,0,1,1);
}

template<class T>
rmMatrix2d<T>::rmMatrix2d(int dx,int dy)
{
	data = NULL;
	resize(0,0,dx,dy);
}

// Copy constructor
template<class T>
rmMatrix2d<T>::rmMatrix2d(rmMatrix2d<T> &im1)
{
	int i,j;
	
	data=NULL;		
	dimX=im1.dimX;dimY=im1.dimY;
	orgX=im1.orgX;orgY=im1.orgY;
	allocateMatrix();
  for (j=0;j<dimY;j++)
		for(i=0;i<dimX;i++)
   		data[(i-orgX)+(dimX-orgX)*(j-orgY)] = im1(i,j);
}


template<class T>
rmMatrix2d<T>::rmMatrix2d(int ox, int oy, int dx,int dy)
{
	data = NULL;
	resize(ox,oy,dx,dy);
}

template<class T>
const T & rmMatrix2d<T>::operator() (int x, int y) const
{
	if ((x<orgX) || (x>=dimX) || (y<orgY) || (y>=dimY))
  {
		cerr<<"Error rmMatrix2d: Out of bounds "<<x << " "<< y<<endl;
		exit(-1);
	}
	else
	return data[(x-orgX)+(dimX-orgX)*(y-orgY)];
}

template<class T>
T & rmMatrix2d<T>::operator() (int x, int y)
{
	if ((x<orgX) || (x>=dimX) || (y<orgY) || (y>=dimY))
  {
		cerr<<"Error rmMatrix2d: Out of bounds "<<x << " "<< y<<endl;
		exit(-1);
	}
	else
	return data[(x-orgX)+(dimX-orgX)*(y-orgY)];
}

template<class T>
void rmMatrix2d<T>::resize(int ox,int oy,int dx, int dy)
{
	deleteMatrix();
	dimX=dx;
	dimY=dy;
	orgX=ox;
	orgY=oy;
	allocateMatrix();
}

template<class T>
void rmMatrix2d<T>::resize(int dx, int dy)
{
	resize(0,0,dx,dy);
}

template<class T>
void rmMatrix2d<T>::resize(int *d)
{
	resize(0,0,d[0],d[1]);
}


template<class T>
rmMatrix2d<T>::~rmMatrix2d()
{
	deleteMatrix();
}

template<class T>
void rmMatrix2d<T>::deleteMatrix()
{
	dimX=dimY=0;
	orgX=orgY=0;
	delete[] data;
}

template<class T>
void rmMatrix2d<T>::allocateMatrix()
{
   data = new T [(dimX-orgX)*(dimY-orgY)];
}

template<class T>
void  rmMatrix2d<T>::operator =(const rmMatrix2d<T> &im1)
{
	int i,j;

	resize(im1.orgX,im1.orgY,im1.dimX,im1.dimY);
  for (j=0;j<dimY;j++)
		for(i=0;i<dimX;i++)
   		data[(i-orgX)+(dimX-orgX)*(j-orgY)] = im1(i,j);

}

template<class T>
void  rmMatrix2d<T>::clear (T value)
{
	int i,j;
	for (j=0;j<dimY;j++)
		for(i=0;i<dimX;i++)
			data[(i-orgX)+(dimX-orgX)*(j-orgY)] = value;

}


template<class T>
rmMatrix2d<T> operator*(const rmMatrix2d<T> &m0,const rmMatrix2d<T> & m1)
{
 	int i,j,l;
	rmMatrix2d<T> res;

	if (m0.dimX!= m1.dimY) {
		error("Multiplication should be: (mxn) * (pxm) dimensions");
	}
	
	res.resize(m1.dimX,m0.dimY);
	for (i = 0; i < m1.dimX; i++) {
     for (l = 0; l < m0.dimY; l++) {
       	res(i,l) = 0;
				for (j = 0; j < m0.dimX; j++)
          res(i,l) += m1(i,j) * m0(j,l);
     }
	}
	return res;
}

template<class T>
rmMatrix2d<T> operator+(const rmMatrix2d<T> &l,const rmMatrix2d<T> &m)

{
	int i,j;
	rmMatrix2d<T> res;

	if ((l.dimX != m.dimX) || (l.dimY!= m.dimY)) {
		error("Sum: matrix should be same dimensions");
	}
	res.resize(l.dimX,l.dimY);	
	for (j=0;j<l.dimY;j++)
		for(i=0;i<l.dimX;i++)
			res(i,j) = l(i,j) + m(i,j);
	return res;
}

template<class T>
rmMatrix2d<T> operator-(const rmMatrix2d<T> &l,const rmMatrix2d<T> &m)
{
	int i,j;
	rmMatrix2d<T> res;

	if ((l.dimX != m.dimX) || (l.dimY!= m.dimY)) {
		error("Subtract: matrix should be same dimensions");
	}
	res.resize(l.dimX,l.dimY);	
	for (j=0;j<l.dimY;j++)
		for(i=0;i<l.dimX;i++)
			res(i,j) = l(i,j) - m(i,j);
	return res;
}


template<class T>
rmMatrix2d<T> rmMatrix2d<T>::transpose()
{
	int i,j;
	rmMatrix2d<T> res;
	
	res.resize(dimY,dimX);
	for (j=0;j<dimY;j++)
		for(i=0;i<dimX;i++)
			res(j,i) = this->operator()(i,j);
	return res;
}
	
template<class T>
rmVector<T> rmMatrix2d<T>::getRow(int r)
{
	int i;
	rmVector<T> res;
	
	res.resize(dimX);
	for(i=0;i<dimX;i++)
		res(i) = this->operator()(i,r);
	return res;
}
	
template<class T>
rmVector<T> rmMatrix2d<T>::getCol(int c)
{
	int j;
	rmVector<T> res;
	
	res.resize(dimY);
	for (j=0;j<dimY;j++)
		res(j) = this->operator()(c,j);
	return res;
}

template<class T>
void rmMatrix2d<T>::setCol(int c,const rmVector<T> &res)
{
	int j;

	if (dimY!= res.dimX) error("Column vector and Matrix of different sizes");
	for (j=0;j<dimY;j++)
		this->operator()(c,j) = res(j);
}

template<class T>
void rmMatrix2d<T>::setRow(int r,const rmVector<T> &res)
{
	int i,j;

	if (dimX!= res.dimX) error("Row vector and Matrix of different sizes");
	for(i=0;i<dimX;i++)
		this->operator()(i,r) = res(i);
}	


template<class T>
void rmMatrix2d<T>::printInfo()
{
	cout << "Matrix dimensions "<< dimX <<" "<< dimY;
	cout << " Orgs "<< orgX <<" "<< orgY << endl;
}

// Class used below for inverse...
class MatxCofactor : public rmMatrix2d<float> {
public:
MatxCofactor (const rmMatrix2d<float> &a,int aI, int aJ, int order)
{
	resize(order,order);
	for (int i=0, k=0; i < a.dimX; ++i) {
		if (i != aI) {
			for (int j=0, l=0; j<a.dimY; ++j) {
				if (j != aJ) {
					this->operator()(k,l) = a(i,j);
					++l;
				}
			}
			++k;
		}
	}
	//cout << "COFACTOR "<< *this;
}
};

template <class T>
rmMatrix2d<T> rmMatrix2d<T>::inverse()
{
	rmMatrix2d<T> b;
	float det;
	b.resize(dimX,dimY);
  for (int i=0; i<dimX; ++i)
		for (int j=0; j<dimY; ++j) {
			int sgn = ( (i+j)%2) ? -1 : 1;
			b(i,j) = sgn * MatxCofactor( *this, i,j,dimX-1).determinant();
	}
	b=b.transpose();
	det = determinant();
	cout << "det " << det << endl;

	for (int i=0; i<dimX; ++i)
		for (int j=0; j<dimY; ++j) {
			b(i,j)=b(i,j) / det;
	}
	
	return b;
}

template <class T>
double rmMatrix2d<T>::determinant()
{
	double d = 0;
	for ( int i = 0; i < dimX; ++i) {
		int sgn = ( i % 2) ? -1 : 1;
		if (dimX>1) {
			MatxCofactor cf(*this,i, 0, dimX-1);
			d += sgn * this->operator()(i,0) * cf.determinant();
		} else d += sgn * this->operator()(i,0);
	}
//	cout <<" d: "<< d<< endl;
//	cout <<" this" << *this;
	return d;
}

template<class T>
ostream& operator<< (ostream& os, const rmMatrix2d<T>& m)
{
	os << "Contents of Matrix:    " << endl;
	for (int j=0;j<m.dimY;j++) {
		for(int i=0;i<m.dimX;i++) {
			os << m(i,j) << "\t";
		}
		os << endl;
	}
	return os;
}
#endif
